#include <cstdio>
#include <iostream>
#include <memory>
#include <cstdlib>

#ifdef WIN32
#include <windows.h> //for Sleep()
#endif


using std::auto_ptr;

#ifndef RTI_USES_STD_FSTREAM
#define RTI_USES_STD_FSTREAM
#endif
#include <RTI.hh>
#include <NullFederateAmbassador.hh>

RTI::ObjectClassHandle	base_id;
RTI::ObjectClassHandle	aircraft_id;
RTI::ObjectClassHandle	submarine_id;	//I know about this, Application1 does not

RTI::ObjectHandle		object_ac_id;
RTI::ObjectHandle		object_sub_id;

//BaseEntity attrs:
RTI::AttributeHandle	attr_location_id;
RTI::AttributeHandle	attr_orientation_id;
RTI::AttributeHandle	attr_velocity_id;

//Aircraft attrs:
RTI::AttributeHandle	attr_afterburner_id;

//Submarine attrs:
RTI::AttributeHandle	attr_periscopeup_id;


class MyFederate: public NullFederateAmbassador
{
	public:
		MyFederate(){};

		void discoverObjectInstance(RTI::ObjectHandle theObject,RTI::ObjectClassHandle theObjectClass,const char *theObjectName) throw ( RTI::CouldNotDiscover,RTI::ObjectClassNotKnown,RTI::FederateInternalError)
		{}

		void reflectAttributeValues(RTI::ObjectHandle theObject,const RTI::AttributeHandleValuePairSet &theAttributes,const char *theTag) throw ( RTI::ObjectNotKnown,RTI::AttributeNotKnown,RTI::FederateOwnsAttributes,RTI::FederateInternalError){}
		
		void removeObjectInstance(RTI::ObjectHandle theObject, const char *tag) throw (RTI::ObjectNotKnown, RTI::FederateInternalError){}
		
		void receiveInteraction(RTI::InteractionClassHandle theInteraction, const RTI::ParameterHandleValuePairSet &theData, const char *theTag) throw (RTI::InteractionClassNotKnown, RTI::InteractionParameterNotKnown, RTI::FederateInternalError){}
};

RTI::RTIambassador		*rtiAmb;
MyFederate				*myfed;

void Snooze(float secs)
{
#ifdef WIN32
	Sleep((int)(secs*1000));
#else
	sleep(secs);
#endif
}

int main()
{
	try
	{
		rtiAmb=new RTI::RTIambassador();
		myfed=new MyFederate();
	}
	catch(RTI::Exception &e)
	{
	  printf("%s--<%s>\n",e._name, e._reason);
		exit(1);
	}

	try
	{
		rtiAmb->createFederationExecution("Recon", "recon.fed");
	}
	catch(RTI::FederationExecutionAlreadyExists &e)
	{
		e._name; //ignore this 'exceptional circumstance' of the federation already existing...
	}
	catch(RTI::Exception &e)
	{
	  printf("%s--<%s>\n",e._name,e._reason);
	  exit(1);
	}

	try
	{
		rtiAmb->joinFederationExecution("App2","Recon",myfed);
	}
	catch(RTI::Exception &e)
	{
	  printf("%s--<%s>\n",e._name,e._reason);
		exit(1);
	}

	base_id = rtiAmb->getObjectClassHandle("BaseEntity");
	aircraft_id = rtiAmb->getObjectClassHandle("Aircraft");
	submarine_id = rtiAmb->getObjectClassHandle("Submarine");

	//base entity attrs
	std::auto_ptr<RTI::AttributeHandleSet> attrs_base(RTI::AttributeHandleSetFactory::create(3));
	attrs_base->add(attr_location_id = rtiAmb->getAttributeHandle("WorldLocation", base_id));
	attrs_base->add(attr_orientation_id = rtiAmb->getAttributeHandle("Orientation", base_id));
	attrs_base->add(attr_velocity_id = rtiAmb->getAttributeHandle("VelocityVector", base_id));
	rtiAmb->subscribeObjectClassAttributes(base_id, *attrs_base);

	//aircraft entity attrs
	std::auto_ptr<RTI::AttributeHandleSet> attrs_ac(RTI::AttributeHandleSetFactory::create(1));
	attrs_ac->add(attr_afterburner_id = rtiAmb->getAttributeHandle("AfterburnerOn", aircraft_id));
	try
	{
		//get error here when running for the second time
		rtiAmb->subscribeObjectClassAttributes(aircraft_id, *attrs_ac);
	}
	catch(RTI::Exception &e)
	{
	  printf("%s--<%s>\n",e._name,e._reason);
		exit(1);
	}

	//submarine entity attrs
	std::auto_ptr<RTI::AttributeHandleSet> attrs_sub(RTI::AttributeHandleSetFactory::create(1));
	attrs_sub->add(attr_periscopeup_id = rtiAmb->getAttributeHandle("PeriscopeUp", submarine_id));
	rtiAmb->subscribeObjectClassAttributes(submarine_id, *attrs_sub);

	//make an aircraft
	std::auto_ptr<RTI::AttributeHandleSet> attrs_myac(RTI::AttributeHandleSetFactory::create(4));
	attrs_myac->add(attr_location_id);
	attrs_myac->add(attr_orientation_id);
	attrs_myac->add(attr_velocity_id);
	attrs_myac->add(attr_afterburner_id);
	rtiAmb->publishObjectClass(aircraft_id, *attrs_myac);
	object_ac_id = rtiAmb->registerObjectInstance(aircraft_id, "Make Aircraft");
	puts("Made Aircraft");

	//make a submarine
	std::auto_ptr<RTI::AttributeHandleSet> attrs_mysub(RTI::AttributeHandleSetFactory::create(4));
	attrs_mysub->add(attr_location_id);
	attrs_mysub->add(attr_orientation_id);
	attrs_mysub->add(attr_velocity_id);
	attrs_mysub->add(attr_periscopeup_id);
	rtiAmb->publishObjectClass(submarine_id, *attrs_mysub);
	object_sub_id = rtiAmb->registerObjectInstance(submarine_id, "Make Submarine");
	puts("Made Submarine");

	printf("Application2: Configured ok now running\n5 dots: ");
	fflush(stdout);

	//now run for 5 seconds
	for(int i=0;i<5;++i)
	{
		Snooze(1.0f);
		printf(".");
		fflush(stdout);
		try
		{
			rtiAmb->tick();
		}
		catch(RTI::Exception &e)
		{
			printf("%s\n",e._name);
			Snooze(2.0f);
			exit(1);
		}
	}
	printf("\nApplication2: now exiting\n");
	
	//now delete objects and resign
	rtiAmb->deleteObjectInstance(object_ac_id,"Delete Aircraft");
	rtiAmb->unpublishObjectClass(aircraft_id);

	rtiAmb->deleteObjectInstance(object_sub_id,"Delete Submarine");
	rtiAmb->unpublishObjectClass(submarine_id);

	rtiAmb->unsubscribeObjectClass(base_id);
	rtiAmb->unsubscribeObjectClass(aircraft_id);
	rtiAmb->unsubscribeObjectClass(submarine_id);
	rtiAmb->resignFederationExecution(RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES);
	try
	{
		rtiAmb->destroyFederationExecution("Recon");//shouldn't work because Application1 is still running
	}
	catch(RTI::FederatesCurrentlyJoined &e)
	{
		e._name;
		//ignore
	}
	catch(RTI::Exception &e)
	{
		printf("%s\n",e._name);
		Snooze(2.0f);
		exit(1);
	}
	//we're done, exit
	try
	{
		delete myfed;
		delete rtiAmb;
	}
	catch(RTI::Exception &e)
	{
		printf("%s\n",e._name);
		
		exit(1);
	}
	puts("Done");
	Snooze(3.0f);
}
